@@ -27,11 +27,15 @@
 #include <sys/stat.h>
 #include <sys/types.h>
 #include <sys/wait.h>
+#include <unistd.h>
+
 #include <glib-unix.h>
 #include <gio/gio.h>
 #include <gio/gunixfdlist.h>
 
 #include "backport-autoptr.h"
+#include "flatpak-portal.h"
+#include "flatpak-session-helper.h"
 
 /* Change to #if 1 to check backwards-compatibility code paths */
 #if 0
@@ -39,34 +43,6 @@
 #define GLIB_CHECK_VERSION(x, y, z) (0)
 #endif
 
-typedef enum {
-  FLATPAK_SPAWN_FLAGS_CLEAR_ENV = 1 << 0,
-  FLATPAK_SPAWN_FLAGS_LATEST_VERSION = 1 << 1,
-  FLATPAK_SPAWN_FLAGS_SANDBOX = 1 << 2,
-  FLATPAK_SPAWN_FLAGS_NO_NETWORK = 1 << 3,
-  FLATPAK_SPAWN_FLAGS_WATCH_BUS = 1 << 4, /* Since 1.2 */
-  FLATPAK_SPAWN_FLAGS_EXPOSE_PIDS = 1 << 5, /* Since 1.6, optional */
-} FlatpakSpawnFlags;
-
-typedef enum {
-  FLATPAK_HOST_COMMAND_FLAGS_CLEAR_ENV = 1 << 0,
-  FLATPAK_HOST_COMMAND_FLAGS_WATCH_BUS = 1 << 1, /* Since 1.2 */
-} FlatpakHostCommandFlags;
-
-/* Since 1.6 */
-typedef enum {
-  FLATPAK_SPAWN_SANDBOX_FLAGS_SHARE_DISPLAY = 1 << 0,
-  FLATPAK_SPAWN_SANDBOX_FLAGS_SHARE_SOUND = 1 << 1,
-  FLATPAK_SPAWN_SANDBOX_FLAGS_SHARE_GPU = 1 << 2,
-  FLATPAK_SPAWN_SANDBOX_FLAGS_ALLOW_DBUS = 1 << 3,
-  FLATPAK_SPAWN_SANDBOX_FLAGS_ALLOW_A11Y = 1 << 4,
-} FlatpakSpawnSandboxFlags;
-
-/* Since 1.6 */
-typedef enum {
-  FLATPAK_SPAWN_SUPPORT_FLAGS_EXPOSE_PIDS = 1 << 0,
-} FlatpakSpawnSupportFlags;
-
 static GDBusConnection *session_bus = NULL;
 
 guint child_pid = 0;
@@ -478,6 +454,17 @@ get_portal_supports (void)
   return supports;
 }
 
+#define NOT_SETUID_ROOT_MESSAGE \
+"This feature requires Flatpak to be using a bubblewrap (bwrap) executable\n" \
+"that is not setuid root.\n" \
+"\n" \
+"The non-setuid version of bubblewrap requires a kernel that allows\n" \
+"unprivileged users to create new user namespaces.\n" \
+"\n" \
+"For more details please see:\n" \
+"https://github.com/flatpak/flatpak/wiki/User-namespace-requirements\n" \
+"\n"
+
 static void
 check_portal_supports (const char *option, guint32 supports_needed)
 {
@@ -486,12 +473,143 @@ check_portal_supports (const char *option, guint32 supports_needed)
   if ((supports & supports_needed) != supports_needed)
     {
       g_printerr ("--%s not supported by host portal\n", option);
+
+      if (supports_needed == FLATPAK_SPAWN_SUPPORT_FLAGS_EXPOSE_PIDS)
+        g_printerr ("\n%s", NOT_SETUID_ROOT_MESSAGE);
+
       exit (1);
     }
 }
 
+/*
+ * @str: A path
+ * @prefix: A possible prefix
+ *
+ * The same as flatpak_has_path_prefix(), but instead of a boolean,
+ * return the part of @str after @prefix (non-%NULL but possibly empty)
+ * if @str has prefix @prefix, or %NULL if it does not.
+ *
+ * Returns: (nullable) (transfer none): the part of @str after @prefix,
+ *  or %NULL if @str is not below @prefix
+ */
+static const char *
+get_path_after (const char *str,
+                const char *prefix)
+{
+  while (TRUE)
+    {
+      /* Skip consecutive slashes to reach next path
+         element */
+      while (*str == '/')
+        str++;
+      while (*prefix == '/')
+        prefix++;
+
+      /* No more prefix path elements? Done! */
+      if (*prefix == 0)
+        return str;
+
+      /* Compare path element */
+      while (*prefix != 0 && *prefix != '/')
+        {
+          if (*str != *prefix)
+            return NULL;
+          str++;
+          prefix++;
+        }
+
+      /* Matched prefix path element,
+         must be entire str path element */
+      if (*str != '/' && *str != 0)
+        return NULL;
+    }
+}
+
+static gint32
+path_to_handle (GUnixFDList *fd_list,
+                const char *path,
+                const char *home_realpath,
+                const char *flatpak_id,
+                GError **error)
+{
+  int path_fd = open (path, O_PATH|O_CLOEXEC|O_NOFOLLOW|O_RDONLY);
+  int saved_errno;
+  gint32 handle;
+
+  if (path_fd < 0)
+    {
+      saved_errno = errno;
+      g_set_error (error, G_IO_ERROR, g_io_error_from_errno (saved_errno),
+                   "Failed to open %s to expose in sandbox: %s",
+                   path, g_strerror (saved_errno));
+      return -1;
+    }
+
+  if (home_realpath != NULL && flatpak_id != NULL)
+    {
+      g_autofree char *real = NULL;
+      const char *after = NULL;
+
+      real = realpath (path, NULL);
+
+      if (real != NULL)
+        after = get_path_after (real, home_realpath);
+
+      if (after != NULL)
+        {
+          g_autofree char *var_path = NULL;
+          int var_fd = -1;
+          struct stat path_buf;
+          struct stat var_buf;
+
+          /* @after is possibly "", but that's OK: if @path is exactly $HOME,
+           * we want to check whether it's the same file as
+           * ~/.var/app/$FLATPAK_ID, with no suffix
+           */
+          var_path = g_build_filename (home_realpath, ".var", "app", flatpak_id,
+                                       after, NULL);
+
+          var_fd = open (var_path, O_PATH|O_CLOEXEC|O_NOFOLLOW|O_RDONLY);
+
+          if (var_fd >= 0 &&
+              fstat (path_fd, &path_buf) == 0 &&
+              fstat (var_fd, &var_buf) == 0 &&
+              path_buf.st_dev == var_buf.st_dev &&
+              path_buf.st_ino == var_buf.st_ino)
+            {
+              close (path_fd);
+              path_fd = var_fd;
+              var_fd = -1;
+            }
+          else
+            {
+              close (var_fd);
+            }
+        }
+    }
+
+
+  handle = g_unix_fd_list_append (fd_list, path_fd, error);
+
+  if (handle < 0)
+    {
+      g_prefix_error (error, "Failed to add fd to list for %s: ", path);
+      close (path_fd);
+      return -1;
+    }
+
+  /* The GUnixFdList keeps a duplicate, so we should release the original */
+  close (path_fd);
+  return handle;
+}
+
 static gboolean
-add_paths_to_variant (GVariantBuilder *builder, GUnixFDList *fd_list, const GStrv paths, gboolean ignore_errors)
+add_paths_to_variant (GVariantBuilder *builder,
+                      GUnixFDList *fd_list,
+                      const GStrv paths,
+                      const char *home_realpath,
+                      const char *flatpak_id,
+                      gboolean ignore_errors)
 {
   g_autoptr(GError) error = NULL;
 
@@ -500,34 +618,145 @@ add_paths_to_variant (GVariantBuilder *builder, GUnixFDList *fd_list, const GStr
 
   for (gsize i = 0; paths[i] != NULL; i++)
     {
-      gint handle = -1;
-      int path_fd = open (paths[i], O_PATH|O_CLOEXEC|O_NOFOLLOW|O_RDONLY);
-      if (path_fd == -1)
+      gint32 handle = path_to_handle (fd_list, paths[i], home_realpath,
+                                      flatpak_id,
+                                      ignore_errors ? NULL : &error);
+
+      if (handle < 0)
         {
           if (ignore_errors)
             continue;
 
-          g_printerr ("Failed to open %s to expose in sandbox\n", paths[i]);
+          g_printerr ("%s\n", error->message);
           return FALSE;
         }
 
-      handle = g_unix_fd_list_append (fd_list, path_fd, &error);
-      if (handle == -1)
-        {
-          if (ignore_errors)
-            continue;
-
-          g_printerr ("Failed to add fd to list for %s: %s\n", paths[i], error->message);
-          return FALSE;
-        }
-      /* The GUnixFdList keeps a duplicate, so we should release the original */
-      close (path_fd);
       g_variant_builder_add (builder, "h", handle);
     }
 
   return TRUE;
 }
 
+static GHashTable *opt_env = NULL;
+static GHashTable *opt_unsetenv = NULL;
+
+static gboolean
+opt_env_cb (G_GNUC_UNUSED const char *option_name,
+            const gchar *value,
+            G_GNUC_UNUSED gpointer data,
+            GError **error)
+{
+  g_auto(GStrv) split = g_strsplit (value, "=", 2);
+
+  g_assert (opt_env != NULL);
+  g_assert (opt_unsetenv != NULL);
+
+  if (split == NULL ||
+      split[0] == NULL ||
+      split[0][0] == 0 ||
+      split[1] == NULL)
+    {
+      g_set_error (error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE,
+                   "Invalid env format %s", value);
+      return FALSE;
+    }
+
+  g_hash_table_remove (opt_unsetenv, split[0]);
+  g_hash_table_replace (opt_env,
+                        g_steal_pointer (&split[0]),
+                        g_steal_pointer (&split[1]));
+  return TRUE;
+}
+
+static gboolean
+opt_unset_env_cb (G_GNUC_UNUSED const char *option_name,
+                  const gchar *value,
+                  G_GNUC_UNUSED gpointer data,
+                  G_GNUC_UNUSED GError **error)
+{
+  g_assert (opt_env != NULL);
+  g_assert (opt_unsetenv != NULL);
+
+  g_hash_table_remove (opt_env, value);
+  g_hash_table_add (opt_unsetenv, g_strdup (value));
+  return TRUE;
+}
+
+static gboolean
+option_env_fd_cb (G_GNUC_UNUSED const gchar *option_name,
+                  const gchar *value,
+                  G_GNUC_UNUSED gpointer data,
+                  GError **error)
+{
+  g_autofree gchar *proc_filename = NULL;
+  g_autofree gchar *env_block = NULL;
+  gsize remaining;
+  const char *p;
+  guint64 fd;
+  gchar *endptr;
+
+  g_assert (opt_env != NULL);
+  g_assert (opt_unsetenv != NULL);
+
+  fd = g_ascii_strtoull (value, &endptr, 10);
+
+  if (endptr == NULL || *endptr != '\0' || fd > G_MAXINT)
+    {
+      g_set_error (error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE,
+                   "Not a valid file descriptor: %s", value);
+      return FALSE;
+    }
+
+  proc_filename = g_strdup_printf ("/proc/self/fd/%d", (int) fd);
+
+  if (!g_file_get_contents (proc_filename, &env_block, &remaining, error))
+    return FALSE;
+
+  p = env_block;
+
+  while (remaining > 0)
+    {
+      g_autofree gchar *var = NULL;
+      g_autofree gchar *val = NULL;
+      size_t len = strnlen (p, remaining);
+      const char *equals;
+
+      g_assert (len <= remaining);
+
+      equals = memchr (p, '=', len);
+
+      if (equals == NULL || equals == p)
+        {
+          g_set_error (error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE,
+                       "Environment variable must be given in the form VARIABLE=VALUE, not %.*s",
+                       (int) len, p);
+          return FALSE;
+        }
+
+      var = g_strndup (p, equals - p);
+      val = g_strndup (equals + 1, len - (equals - p) - 1);
+      g_hash_table_remove (opt_unsetenv, var);
+      g_hash_table_replace (opt_env,
+                            g_steal_pointer (&var),
+                            g_steal_pointer (&val));
+
+      p += len;
+      remaining -= len;
+
+      if (remaining > 0)
+        {
+          g_assert (*p == '\0');
+          p += 1;
+          remaining -= 1;
+        }
+    }
+
+  if (fd >= 3)
+    close (fd);
+
+  return TRUE;
+}
+
 int
 main (int    argc,
       char **argv)
@@ -539,11 +768,11 @@ main (int    argc,
   int i, opt_argc;
   gboolean verbose = FALSE;
   char **forward_fds = NULL;
-  char **opt_envs = NULL;
   guint spawn_flags;
   gboolean opt_clear_env = FALSE;
   gboolean opt_watch_bus = FALSE;
   gboolean opt_expose_pids = FALSE;
+  gboolean opt_share_pids = FALSE;
   gboolean opt_latest_version = FALSE;
   gboolean opt_sandbox = FALSE;
   gboolean opt_no_network = FALSE;
@@ -554,7 +783,11 @@ main (int    argc,
   char **opt_sandbox_expose_path_try = NULL;
   char **opt_sandbox_expose_path_ro_try = NULL;
   char *opt_directory = NULL;
+  char *opt_app_path = NULL;
+  char *opt_usr_path = NULL;
   g_autofree char *cwd = NULL;
+  g_autofree char *home_realpath = NULL;
+  const char *flatpak_id = NULL;
   GVariantBuilder options_builder;
   const GOptionEntry options[] = {
     { "verbose", 'v', 0, G_OPTION_ARG_NONE, &verbose,  "Enable debug output", NULL },
@@ -562,7 +795,10 @@ main (int    argc,
     { "clear-env", 0, 0, G_OPTION_ARG_NONE, &opt_clear_env,  "Run with clean environment", NULL },
     { "watch-bus", 0, 0, G_OPTION_ARG_NONE, &opt_watch_bus,  "Make the spawned command exit if we do", NULL },
     { "expose-pids", 0, 0, G_OPTION_ARG_NONE, &opt_expose_pids, "Expose sandbox pid in calling sandbox", NULL },
-    { "env", 0, 0, G_OPTION_ARG_STRING_ARRAY, &opt_envs, "Set environment variable", "VAR=VALUE" },
+    { "share-pids", 0, 0, G_OPTION_ARG_NONE, &opt_share_pids, "Use same pid namespace as calling sandbox", NULL },
+    { "env", 0, 0, G_OPTION_ARG_CALLBACK, &opt_env_cb, "Set environment variable", "VAR=VALUE" },
+    { "unset-env", 0, 0, G_OPTION_ARG_CALLBACK, &opt_unset_env_cb, "Unset environment variable", "VAR=VALUE" },
+    { "env-fd", 0, 0, G_OPTION_ARG_CALLBACK, &option_env_fd_cb, "Read environment variables in env -0 format from FD", "FD" },
     { "latest-version", 0, 0, G_OPTION_ARG_NONE, &opt_latest_version,  "Run latest version", NULL },
     { "sandbox", 0, 0, G_OPTION_ARG_NONE, &opt_sandbox,  "Run sandboxed", NULL },
     { "no-network", 0, 0, G_OPTION_ARG_NONE, &opt_no_network,  "Run without network access", NULL },
@@ -575,9 +811,13 @@ main (int    argc,
     { "sandbox-flag", 0, 0, G_OPTION_ARG_CALLBACK, sandbox_flag_callback, "Enable sandbox flag", "FLAG" },
     { "host", 0, 0, G_OPTION_ARG_NONE, &opt_host, "Start the command on the host", NULL },
     { "directory", 0, 0, G_OPTION_ARG_FILENAME, &opt_directory, "Working directory in which to run the command", "DIR" },
+    { "app-path", 0, 0, G_OPTION_ARG_FILENAME, &opt_app_path, "Replace runtime's /app with DIR or empty", "DIR|\"\"" },
+    { "usr-path", 0, 0, G_OPTION_ARG_FILENAME, &opt_usr_path, "Replace runtime's /usr with DIR", "DIR" },
     { NULL }
   };
   guint signal_source = 0;
+  GHashTableIter iter;
+  gpointer key, value;
 
   setlocale (LC_ALL, "");
 
@@ -594,6 +834,8 @@ main (int    argc,
     i++;
 
   opt_argc = i;
+  opt_env = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);
+  opt_unsetenv = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
 
   while (i < argc)
     {
@@ -632,6 +874,11 @@ main (int    argc,
   if (signal_source == 0)
     return 1;
 
+  flatpak_id = g_getenv ("FLATPAK_ID");
+
+  if (flatpak_id != NULL)
+    home_realpath = realpath (g_get_home_dir (), NULL);
+
   session_bus = g_bus_get_sync (G_BUS_TYPE_SESSION, NULL, &error);
   if (session_bus == NULL)
     {
@@ -641,15 +888,15 @@ main (int    argc,
 
   if (opt_host)
     {
-      service_iface = "org.freedesktop.Flatpak.Development";
-      service_obj_path = "/org/freedesktop/Flatpak/Development";
-      service_bus_name = "org.freedesktop.Flatpak";
+      service_iface = FLATPAK_SESSION_HELPER_INTERFACE_DEVELOPMENT;
+      service_obj_path = FLATPAK_SESSION_HELPER_PATH_DEVELOPMENT;
+      service_bus_name = FLATPAK_SESSION_HELPER_BUS_NAME;
     }
   else
     {
-      service_iface = "org.freedesktop.portal.Flatpak";
-      service_obj_path = "/org/freedesktop/portal/Flatpak";
-      service_bus_name = "org.freedesktop.portal.Flatpak";
+      service_iface = FLATPAK_PORTAL_INTERFACE;
+      service_obj_path = FLATPAK_PORTAL_PATH;
+      service_bus_name = FLATPAK_PORTAL_BUS_NAME;
     }
 
   g_dbus_connection_signal_subscribe (session_bus,
@@ -718,18 +965,12 @@ main (int    argc,
       g_variant_builder_add (fd_builder, "{uh}", fd, handle);
     }
 
-  for (i = 0; opt_envs != NULL && opt_envs[i] != NULL; i++)
-    {
-      const char *opt_env = opt_envs[i];
-      g_auto(GStrv) split = g_strsplit (opt_env, "=", 2);
+  g_hash_table_iter_init (&iter, opt_env);
 
-      if (split == NULL || split[0] == NULL || split[0][0] == 0 || split[1] == NULL)
-        {
-          g_printerr ("Invalid env format %s\n", opt_env);
-          return 1;
-        }
-      g_variant_builder_add (env_builder, "{ss}", split[0], split[1]);
-    }
+  while (g_hash_table_iter_next (&iter, &key, &value))
+    g_variant_builder_add (env_builder, "{ss}", key, value);
+
+  g_clear_pointer (&opt_env, g_hash_table_unref);
 
   spawn_flags = 0;
 
@@ -739,11 +980,24 @@ main (int    argc,
   if (opt_watch_bus)
     spawn_flags |= opt_host ? FLATPAK_HOST_COMMAND_FLAGS_WATCH_BUS : FLATPAK_SPAWN_FLAGS_WATCH_BUS;
 
-  if (opt_expose_pids)
+  if (opt_share_pids)
     {
       if (opt_host)
         {
-          g_printerr ("--host not compatible with --expose-pids");
+          g_printerr ("--host not compatible with --share-pids\n");
+          return 1;
+        }
+
+      check_portal_version ("share-pids", 5);
+      check_portal_supports ("share-pids", FLATPAK_SPAWN_SUPPORT_FLAGS_SHARE_PIDS);
+
+      spawn_flags |= FLATPAK_SPAWN_FLAGS_SHARE_PIDS;
+    }
+  else if (opt_expose_pids)
+    {
+      if (opt_host)
+        {
+          g_printerr ("--host not compatible with --expose-pids\n");
           return 1;
         }
 
@@ -757,7 +1011,7 @@ main (int    argc,
     {
       if (opt_host)
         {
-          g_printerr ("--host not compatible with --latest-version");
+          g_printerr ("--host not compatible with --latest-version\n");
           return 1;
         }
       spawn_flags |= FLATPAK_SPAWN_FLAGS_LATEST_VERSION;
@@ -767,7 +1021,7 @@ main (int    argc,
     {
       if (opt_host)
         {
-          g_printerr ("--host not compatible with --sandbox");
+          g_printerr ("--host not compatible with --sandbox\n");
           return 1;
         }
       spawn_flags |= FLATPAK_SPAWN_FLAGS_SANDBOX;
@@ -777,7 +1031,7 @@ main (int    argc,
     {
       if (opt_host)
         {
-          g_printerr ("--host not compatible with --no-network");
+          g_printerr ("--host not compatible with --no-network\n");
           return 1;
         }
       spawn_flags |= FLATPAK_SPAWN_FLAGS_NO_NETWORK;
@@ -785,11 +1039,63 @@ main (int    argc,
 
   g_variant_builder_init (&options_builder, G_VARIANT_TYPE ("a{sv}"));
 
+  if (g_hash_table_size (opt_unsetenv) > 0)
+    {
+      g_hash_table_iter_init (&iter, opt_unsetenv);
+
+      /* The host portal doesn't support options, so we always have to do
+       * this the hard way. The subsandbox portal supports unset-env in
+       * versions >= 5. */
+      if (opt_host ? FALSE : (get_portal_version () >= 5))
+        {
+          GVariantBuilder strv_builder;
+
+          g_variant_builder_init (&strv_builder, G_VARIANT_TYPE_STRING_ARRAY);
+
+          while (g_hash_table_iter_next (&iter, &key, NULL))
+            g_variant_builder_add (&strv_builder, "s", key);
+
+          g_variant_builder_add (&options_builder, "{s@v}", "unset-env",
+                                 g_variant_new_variant (g_variant_builder_end (&strv_builder)));
+        }
+      else
+        {
+          /* env(1) will do the wrong thing if argv[0] contains an equals
+           * sign, so we might need to prepend this incantation - and
+           * because we're prepending, we need to do it backwards.
+           * More legibly, we're replacing MY=COMMAND ARGS with:
+           *
+           *     /usr/bin/env -u VAR -u VAR2 /bin/sh -euc 'exec "$@"' sh MY=COMMAND ARGS
+           *
+           * This is a standard trick for dealing with env(1). */
+          g_assert (child_argv->len >= 1);
+
+          if (strchr (g_ptr_array_index (child_argv, 0), '=') != NULL)
+            {
+              g_ptr_array_insert (child_argv, 0, g_strdup ("sh"));  /* argv[0] */
+              g_ptr_array_insert (child_argv, 0, g_strdup ("exec \"$@\""));
+              g_ptr_array_insert (child_argv, 0, g_strdup ("-euc"));
+              g_ptr_array_insert (child_argv, 0, g_strdup ("/bin/sh"));
+            }
+
+          while (g_hash_table_iter_next (&iter, &key, NULL))
+            {
+              /* Again, yes, this is backwards: we're prepending. */
+              g_ptr_array_insert (child_argv, 0, g_strdup (key));
+              g_ptr_array_insert (child_argv, 0, g_strdup ("-u"));
+            }
+
+          g_ptr_array_insert (child_argv, 0, g_strdup ("/usr/bin/env"));
+        }
+    }
+
+  g_clear_pointer (&opt_unsetenv, g_hash_table_unref);
+
   if (opt_sandbox_expose)
     {
       if (opt_host)
         {
-          g_printerr ("--host not compatible with --sandbox-expose");
+          g_printerr ("--host not compatible with --sandbox-expose\n");
           return 1;
         }
       g_variant_builder_add (&options_builder, "{s@v}", "sandbox-expose",
@@ -800,7 +1106,7 @@ main (int    argc,
     {
       if (opt_host)
         {
-          g_printerr ("--host not compatible with --sandbox-expose-ro");
+          g_printerr ("--host not compatible with --sandbox-expose-ro\n");
           return 1;
         }
       g_variant_builder_add (&options_builder, "{s@v}", "sandbox-expose-ro",
@@ -833,8 +1139,8 @@ main (int    argc,
 
       check_portal_version ("sandbox-expose-path", 3);
 
-      if (!add_paths_to_variant (expose_fd_builder, fd_list, opt_sandbox_expose_path, FALSE)
-          || !add_paths_to_variant (expose_fd_builder, fd_list, opt_sandbox_expose_path_try, TRUE))
+      if (!add_paths_to_variant (expose_fd_builder, fd_list, opt_sandbox_expose_path, home_realpath, flatpak_id, FALSE)
+          || !add_paths_to_variant (expose_fd_builder, fd_list, opt_sandbox_expose_path_try, home_realpath, flatpak_id, TRUE))
         return 1;
 
       g_variant_builder_add (&options_builder, "{s@v}", "sandbox-expose-fd",
@@ -853,14 +1159,76 @@ main (int    argc,
 
       check_portal_version ("sandbox-expose-path-ro", 3);
 
-      if (!add_paths_to_variant (expose_fd_builder, fd_list, opt_sandbox_expose_path_ro, FALSE)
-          || !add_paths_to_variant (expose_fd_builder, fd_list, opt_sandbox_expose_path_ro_try, TRUE))
+      if (!add_paths_to_variant (expose_fd_builder, fd_list, opt_sandbox_expose_path_ro, home_realpath, flatpak_id, FALSE)
+          || !add_paths_to_variant (expose_fd_builder, fd_list, opt_sandbox_expose_path_ro_try, home_realpath, flatpak_id, TRUE))
         return 1;
 
       g_variant_builder_add (&options_builder, "{s@v}", "sandbox-expose-fd-ro",
                              g_variant_new_variant (g_variant_builder_end (g_steal_pointer (&expose_fd_builder))));
     }
 
+  if (opt_app_path != NULL)
+    {
+      gint32 handle;
+
+      g_debug ("Using \"%s\" as /app instead of runtime", opt_app_path);
+
+      if (opt_host)
+        {
+          g_printerr ("--host not compatible with --app-path\n");
+          return 1;
+        }
+
+      check_portal_version ("app-path", 6);
+
+      if (opt_app_path[0] == '\0')
+        {
+          /* Empty path is special-cased to mean an empty directory */
+          spawn_flags |= FLATPAK_SPAWN_FLAGS_EMPTY_APP;
+        }
+      else
+        {
+          handle = path_to_handle (fd_list, opt_app_path, home_realpath,
+                                   flatpak_id, &error);
+
+          if (handle < 0)
+            {
+              g_printerr ("%s\n", error->message);
+              return 1;
+            }
+
+          g_variant_builder_add (&options_builder, "{s@v}", "app-fd",
+                                 g_variant_new_variant (g_variant_new_handle (handle)));
+        }
+    }
+
+  if (opt_usr_path != NULL)
+    {
+      gint32 handle;
+
+      g_debug ("Using %s as /usr instead of runtime", opt_usr_path);
+
+      if (opt_host)
+        {
+          g_printerr ("--host not compatible with --usr-path\n");
+          return 1;
+        }
+
+      check_portal_version ("usr-path", 6);
+
+      handle = path_to_handle (fd_list, opt_usr_path, home_realpath,
+                               flatpak_id, &error);
+
+      if (handle < 0)
+        {
+          g_printerr ("%s\n", error->message);
+          return 1;
+        }
+
+      g_variant_builder_add (&options_builder, "{s@v}", "usr-fd",
+                             g_variant_new_variant (g_variant_new_handle (handle)));
+    }
+
   if (!opt_directory)
     {
       opt_directory = cwd;
